// SharedMemoryReader.cpp
// 2016.05 Kyoung Shin Park 
//

#include "QUANTA/QUANTAinit.hxx"
#include "QUANTA/QUANTAnet_datapack_c.hxx"
#include "SharedMemory.h"
#include "Semaphore.h"
#include <windows.h>
#include <iostream>
#include <string.h>

#define BUFFER_SIZE 4096

// shared memory 
CSharedMemory shmem("_SHMEM_TEST");
char * shmbuf;
HANDLE segment_id = (HANDLE)3782;
// semaphore (mutex)
KSemaphore sem("_SEM_TEST");
// datapack
bool datapackActive = false;
static QUANTAnet_datapack_c dataPacker;

void parseCommandLine(int argc, char ** argv)
{
    for (int i = 1; i < argc; i++) {
		if (!strncmp(argv[i], "--", 2)) {
			if (!strncmp(argv[i], "--datapack", 10)) {
				datapackActive = true;
				printf("datapack active\n");
			}
		}
	}
}

// read data from shared memory and unpack data
int unpackDataFrom(char * buffer)
{
	int receivedInt;
	float receivedFloat;
	float* floatArray = new float[4];
	double receivedDouble;
	long receivedLong;
	unsigned long receivedUnsignedLong;
	char receivedChar;
	char receivedStr[8];
	char receivedString[8];

	int dataSize = dataPacker.sizeof_int()	// int
		+ dataPacker.sizeof_float()			// float
		+ dataPacker.sizeof_float(4)		// float array
		+ dataPacker.sizeof_double()		// double
		+ dataPacker.sizeof_long()			// long
		+ dataPacker.sizeof_unsignedLong()	// unsigned long
		+ dataPacker.sizeof_char()			// char
		+ dataPacker.sizeof_char(8)			// char array
		+ dataPacker.sizeof_string(8);		// string

	sem.lock();
	dataPacker.initUnpack(buffer, dataSize);
	dataPacker.unpackInt(&receivedInt);
	dataPacker.unpackFloat(&receivedFloat);
	dataPacker.unpackFloatArray(floatArray, 4);
	dataPacker.unpackDouble(&receivedDouble);
	dataPacker.unpackLong(&receivedLong);
	dataPacker.unpackUnsignedLong(&receivedUnsignedLong);
	dataPacker.unpackChar(&receivedChar);
	dataPacker.unpack(receivedStr, 8);
	dataPacker.unpackString(receivedString, 8);
	printf("Received Int: %d\n", receivedInt);
	printf("Received Float: %f\n", receivedFloat);
	printf("Received Float Array - with the floats: ");
	for (int i = 0; i<4; i++) printf("%f\t", floatArray[i]);
	printf("\nReceived Double: %f\n", receivedDouble);
	printf("Received Long: %ld\n", receivedLong);
	printf("Received Unsigned Long: %lu\n", receivedUnsignedLong);
	printf("Received Char: %c\n", receivedChar);
	printf("Received Char[]: %s\n", receivedStr);
	printf("Received String: %s\n", receivedString);
	sem.unlock();

	return dataSize;
}

int readDataFrom(char* buffer)
{
	sem.lock();
	printf("%s\n", buffer);
	sem.unlock();
	return sizeof(int);
}

int main(int argc, char ** argv)
{
	parseCommandLine(argc, argv);

	// open semaphore using keyname
	sem.open();
	// initialize shared memory unit & allocate shared memory segment
	segment_id = shmem.getShm((key_t)segment_id, BUFFER_SIZE, S_IRUSR | S_IWUSR);
	printf("shmem id=%d\n", (int)segment_id);
	// attach the shared memory segment
	shmbuf = (char *) shmem.attachShm(segment_id, NULL, 0);
	printf("shared memory attached\n");
	while (1) {
		int dataSize;
		if (datapackActive) {
			dataSize = unpackDataFrom(shmbuf);
		}
		else {
			dataSize = readDataFrom(shmbuf);
		}
		printf("Shared Memory Data ߽ϴ: DataSize=%d\n", dataSize);
		Sleep(500);
	}

	// detach the shared memory segment
	shmem.detachShm(shmbuf);
	// remove the shared memory segment
	shmem.cleanShm(segment_id, IPC_RMID, NULL);
	sem.close();
	printf("shared memory detached & cleanup \n");

	return 0;
}

